function get_sprite(path, transparent)
  local s = diretheme.get_scale();
  
  if not menu_sprites[path] then
    if s ~= 1 then
      menu_sprites[path] = pixels.new(path):scale(s, s):sprite();
    else
      menu_sprites[path] = sprite.new(path);
    end;
    if transparent then
      menu_sprites[path]:colorkey(transparent);
    end;
  end;
  
  return menu_sprites[path];
end;

function get_button(button_name)
  local path = 'images/title/' .. button_name .. '.' .. get_language() .. '.png';
  if menu_sprites[path] then
    return menu_sprites[path];
  end;
  
  local sprite = nil;
  
  --Check if file exists(only used here, so not moved to function)
  local f = io.open(path, 'r');  
  if f then
    f:close();
    sprite = get_sprite(path, 'fuchsia')
  end;
  
  if not sprite then
    local path = 'images/title/' .. button_name .. '.' .. get_default_language() .. '.png';
    sprite = get_sprite(path, 'fuchsia');
  end;
  return sprite
end;

function stop_menu()
  D {'bgframe'};
  D {'background'};
  D {'bganim'};
  D {'title'};
end;

function init_menu(with_text, buttons)
  local scale = diretheme.get_scale();
  
  local w = 960 * scale;
  local h = 600 * scale;
  local sw, sh = diretheme.get_screen();
  local ox = math.floor((sw - w) / 2);
  local oy = math.floor((sh - h) / 2)
  
  local button_dy = 0;
  local button_dx = 0;
  if sh < 520 * scale then
    button_dy = -50;
    button_dx = 20;
  end;
  
  local btop = 515;
  
  if #buttons == 5 then
    btop = 435;
  end;
  
  if sh <= 560 * scale then
    btop = btop - 20;
  end;
  
  if #buttons == 1 then
    D {
      buttons[1], 
      'img', 
      get_button(buttons[1]),
      x=ox + 398 * scale, 
      y=oy + (btop + button_dy) * scale,
      click=true,
      background=true,
      z=3
    };
  else
    D {
      buttons[1], 
      'img', 
      get_button(buttons[1]),
      x=ox + (157 + button_dx) * scale, 
      y=oy + (btop + button_dy) * scale,
      click=true,
      background=true,
      z=3
    };
    D {
      buttons[2], 
      'img', 
      get_button(buttons[2]),
      x=ox + 398 * scale,
      y=oy + (btop + button_dy) * scale,
      click=true,
      background=true,
      z=3
    };
    D {
      buttons[3], 
      'img', 
      get_button(buttons[3]),
      x=ox + (639 - button_dx) * scale,
      y=oy + (btop + button_dy) * scale,
      click=true,
      background=true,
      z=3
    };
  end;
  
  if #buttons == 5 then
    btop = btop + 80
    D {
      buttons[4], 
      'img', 
      get_button(buttons[4]),
      x=ox + (278 + button_dx / 2) * scale, 
      y=oy + (btop + button_dy) * scale,
      click=true,
      background=true,
      z=3
    };
    
    D {
      buttons[5], 
      'img', 
      get_button(buttons[5]),
      x=ox + (519 - button_dx / 2) * scale, 
      y=oy + (btop + button_dy) * scale,
      click=true,
      background=true,
      z=3
    };
  end;
  
  if sw > w or sh > h then
    local bgframe_pixels = pixels.new('images/main_theme/text_bg.png');
    local tbg, tbg_slice;
    tbg, tbg_slice = dire_slice_9(bgframe_pixels, {
      x = 9,
      y = 6,
      w = 96,
      h = 96
    });
    
    local bgframe = pixels.new(978, 612);
    dire_draw_9(tbg, bgframe, {x = 0; y = 0; w = 978; h = 612}, {parts=tbg_slice});
    D {'bgframe', 'img', bgframe:scale(scale, scale):sprite(), x=ox - 9 * scale, y=oy - 6 * scale, w=w + 18 * scale, h=h + 12 * scale, background=true, z=5};
  end;
  
  if not D('background') then
    D {'background', 'img', get_sprite('images/title/bg.png'), x=ox, y=oy, z=4, background=true};
  end;

  if prefs.simplified_animation then
    D {'bganim'};
  elseif not D('bganim') then
    D {'bganim', 'img', get_sprite('images/title/bg_frames.png'), x=ox, y=oy, w=w, h=h, frames=16, z=3, delay=200, animated=true, background=true, frame_data=scale_frame_data({
       {},
       {x=0, y=0, w=960, h=302, dx=0, dy=0},
       {x=960, y=0, w=960, h=302, dx=0, dy=0},
       {x=1920, y=0, w=960, h=302, dx=0, dy=0},
       {x=2880, y=0, w=960, h=302, dx=0, dy=0},
       {x=0, y=302, w=960, h=302, dx=0, dy=0},
       {x=960, y=302, w=960, h=302, dx=0, dy=0},
       {x=1920, y=302, w=960, h=302, dx=0, dy=0},
       {x=2880, y=302, w=960, h=302, dx=0, dy=0},
       {x=0, y=604, w=960, h=302, dx=0, dy=0},
       {x=960, y=604, w=960, h=302, dx=0, dy=0},
       {x=1920, y=604, w=960, h=302, dx=0, dy=0},
       {x=2880, y=604, w=960, h=302, dx=0, dy=0},
       {x=0, y=906, w=960, h=302, dx=0, dy=0},
       {x=960, y=906, w=960, h=302, dx=0, dy=0},
       {x=1920, y=906, w=960, h=302, dx=0, dy=0}
    })};
  end;
  if not D('title') then
    D {'title', 'img', get_sprite('images/title/title.png', 'fuchsia'), x=ox + 230 * scale, y=oy + 88  * scale, z=2, background=true};
  end;
  
  if with_text then
    D {'textbg', 'img', get_sprite('images/title/text_background.png', 'fuchsia'), x=ox + 180 * scale, z=1, y=oy + (126 + button_dy) * scale, background=true};
    if sh < 520 * scale then
      D {'title'}
    else
      D('title').y = oy + 40 * scale;
    end;
  else
    D('title').y = oy + 88 * scale;
  end;
  
  timer:set(25);
end;

room {
  nam = 'main';
  notitle = true;
  nosave = true;
  noautosave = true;
  noinv = true;
  nofading = true;
  disp = '';
  theme = 'menu';
  onenter = function(s)
    in_game = false;
  end;
  themeInit = function(s)
    init_menu(false, {'start', 'load', 'about', 'settings', 'quit'});
  end;
  themeDispose = function(s)
    D {'start'};
    D {'load'};
    D {'about'};
    D {'settings'};
    D {'quit'};
    timer:stop();
  end;
  ondecor = function(s, name, press, x, y, btn)
    if press then
      if name == 'start' then
        walk('intro_start');
        return true;
      elseif  name == 'load' then
        instead.menu('load');
        return true;
      elseif name == 'about' then
        walk('intro_info');
        return true;
      elseif name == 'settings' then
        walk('intro_settings');
        return true;
      elseif  name == 'quit' then
        instead.menu('exit');
        return true;
      end;
    end;
    return false;
  end;
  enter = function(s)
    if prefs.completed then
      snd.music('sound/bgm_credits.ogg');
    else
      snd.music('sound/bgm_intro.ogg');
    end;
  end;
}

room {
  nam = 'intro_start';
  notitle = true;
  nosave = true;
  noautosave = true;
  noinv = true;
  nofading = true;
  disp = '';
  theme = 'menu';
  themeInit = function(s)
    init_menu(true, {'back', 'tutorial', 'start'});
  end;
  themeDispose = function(s)
    D {'back'};
    D {'tutorial'};
    D {'start'};
    D {'textbg'};
    timer:stop();
  end;
  ondecor = function(s, name, press, x, y, btn)
    if press then
      if name == 'start' then
        walk('above_forest');
        in_game = true;
        return true;
      elseif name == 'tutorial' then
        walk('tutorial_start');
        in_game = true;
        return true;
      elseif name == 'back' then
        walk('main');
        return true;
      end;
    end;
    return false;
  end;
  decor = translate("menu_locale", 'intro');
}

room {
  nam = 'intro_info';
  notitle = true;
  nosave = true;
  noautosave = true;
  noinv = true;
  nofading = true;
  disp = '';
  theme = 'menu';
  themeInit = function(s)
    init_menu(true, {'back'});
  end;
  themeDispose = function(s)
    D {'start'};
    D {'back'};
    D {'settings'};
    D {'textbg'};
    timer:stop();
  end;
  ondecor = function(s, name, press, x, y, btn)
    if press then
      if name == 'back' then
        if in_game then
          walk('menu');
        else
          walk('main');
        end;
        return true;
      end;
    end;
    return false;
  end;
  decor = function(s)
    return credits_text(false);
  end;
}

room {
  nam = 'intro_settings';
  nofollow = true;
  notitle = true;
  nosave = true;
  noautosave = true;
  noinv = true;
  nofading = true;
  disp = '';
  theme = 'menu';
  themeInit = function(s)
    init_menu(true, {'back'});
  end;
  themeDispose = function(s)
    D {'start'};
    D {'about'};
    D {'back'};
    D {'textbg'};
    timer:stop();
  end;
  ondecor = function(s, name, press, x, y, btn)
    if press then
      if name == 'back' then
        if in_game then
          stop_menu();
          walkback();
        else
          walk('main');
        end;
        return true;
      end;
    end;
    return false;
  end;
}:with {
  obj {
    nam = '#system_settings';
    dsc = function(s)
      return fmt.c(fmt.b('{[' .. translate_now("menu_locale", 'system_settings') .. ']}^^'));
    end;
    act = function(s)
      instead.menu('settings');
    end;
  };
  obj {
    nam = '#language';
    dsc = function(s)
      local l = translate_now("menu_locale", 'auto_option');
      

      pn(fmt.c(fmt.b(translate_now("menu_locale", 'language'))));
      
      local t;
      if get_language_saved() == '' then
        t = l;
      else
        t = '{#l_auto|'.. l .. '}';
      end;
      
      for i, language in ipairs(get_language_list()) do
        if get_language_saved() == language[1] then
          t = t .. ' | ' .. language[2];
        else
          t = t .. ' | ' .. '{@set_language ' .. language[1] .. '|' .. language[2] .. '}';
        end;
      end;
    
      pn(fmt.c(t) .. '^')
    end;
  }: with {
    obj {
      nam = '#l_auto';
      act = function(s)
        set_language('');
        walk(here());
      end;
    };
  };
  obj {
    nam = '#graphics_scale';
    dsc = function(s)
      local graphic_scale = get_graphic_scale();
      
      pn(fmt.c(fmt.b(translate_now("menu_locale", 'graphic_scale')) .. ' {[' .. translate_now("menu_locale", 'auto_option') .. ']}'))
    
      local t = '';
      
      if graphic_scale > 1 then
        t = t .. '{#g_less|<<} ';
      else
        t = t .. '<< ';
      end;

      local l =  translate_now("menu_locale", 'auto_option') .. ' (' .. tostring(graphic_scale * 100) .. '%)';
      
      
      if prefs.graphic_scale then
        l = tostring(graphic_scale * 100) .. '%';
      end;

      t = t .. l;

      local w, h = diretheme.get_screen();
    
      local wsmax = math.floor(w / 640);
      local smax = math.floor(h / 480);
      if wsmax < smax then
        smax = wsmax;
      end;
    
      if smax < 1 then
        smax = 1;
      end;
    
      if graphic_scale < smax then
        t = t .. ' {#g_more|>>}';
      else
        t = t .. ' >>';
      end;
    
      pn(fmt.c(t) .. '^');
    end;
    act = function(s)
      prefs.graphic_scale = false;
      prefs:store();
    
      set_theme(std.call(here(), 'theme'));
      menu_sprites = {};
      here():themeDispose();
      stop_menu();
      here():themeInit();
      walk(here());
    end;
  }: with {
    obj {
      nam = '#g_less';
      act = function(s)
        prefs.graphic_scale = get_graphic_scale() - 1;
        
        prefs:store();
    
        set_theme(std.call(here(), 'theme'));
        menu_sprites = {};
        here():themeDispose();
        stop_menu();
        here():themeInit();
        walk(here());
      end;
    };
    obj {
      nam = '#g_more';
      act = function(s)
        prefs.graphic_scale = get_graphic_scale() + 1;
        
        prefs:store();
    
        set_theme(std.call(here(), 'theme'));
        menu_sprites = {};
        here():themeDispose();
        stop_menu();
        here():themeInit();
        walk(here());
      end;
    };
  };
  obj {
    nam = '#cursor_scale';
    dsc = function(s)
      local cursor_scale = get_cursor_scale();
      
      pn(fmt.c(fmt.b(translate_now("menu_locale", 'cursor_scale')) .. ' {[' .. translate_now("menu_locale", 'auto_option') .. ']}'))
    
      local t = '';
      
      if cursor_scale > 1 then
        t = t .. '{#cursor_less|<<} ';
      else
        t = t .. '<< ';
      end;

      local l = translate_now("menu_locale", 'auto_option') .. ' (' .. tostring(cursor_scale * 100) .. '%)';
      
      if prefs.cursor_scale then
        l = tostring(cursor_scale * 100) .. '%';
      end;

      t = t .. l;

      local w, h = diretheme.get_screen();
    
      local wsmax = math.floor(w / 640);
      local smax = math.floor(h / 480);
      if wsmax < smax then
        smax = wsmax;
      end;
    
      if smax < 1 then
        smax = 1;
      end;
    
      if cursor_scale < smax then
        t = t .. ' {#cursor_more|>>}';
      else
        t = t .. ' >>';
      end;
    
      pn(fmt.c(t) .. '^');
    end;
    act = function(s)
      prefs.cursor_scale = false;
      prefs:store();
    
      set_theme(std.call(here(), 'theme'));
      menu_sprites = {};
      here():themeDispose();
      stop_menu();
      here():themeInit();
      walk(here());
    end;
  }: with {
    obj {
      nam = '#cursor_less';
      act = function(s)
        prefs.cursor_scale = get_cursor_scale() - 1;
        
        prefs:store();
    
        set_theme(std.call(here(), 'theme'));
        menu_sprites = {};
        here():themeDispose();
        stop_menu();
        here():themeInit();
        walk(here());
      end;
    };
    obj {
      nam = '#cursor_more';
      act = function(s)
        prefs.cursor_scale = get_cursor_scale() + 1;
        
        prefs:store();
    
        set_theme(std.call(here(), 'theme'));
        menu_sprites = {};
        here():themeDispose();
        stop_menu();
        here():themeInit();
        walk(here());
      end;
    };
  };
  obj {
    nam = '#font_scale';
    dsc = function(s)    
      local graphic_scale = get_graphic_scale();
      local font_scale = get_font_scale();
        
      pn(fmt.c(fmt.b(translate_now("menu_locale", 'font_scale')) .. ' {[' .. translate_now("menu_locale", 'auto_option') .. ']}'));
    
      local t = '';
    
      if font_scale > 50 then
        t = t .. '{#f_less|<<} ';
      else
        t = t .. '<< ';
      end;
    
      if prefs.font_scale then
        t = t .. tostring(font_scale) .. '%';
      else
        t = t .. translate_now("menu_locale", 'auto_option') .. ' (' .. tostring(font_scale) .. '%)';
      end;
    
    
      if font_scale < graphic_scale * 200 then
        t = t .. ' {#f_more|>>}';
      else
        t = t .. ' >>';
      end;
    
      pn(fmt.c(t) .. '^');
    end;
    act = function(s)
      prefs.font_scale = false;
    
      prefs:store();
      set_theme(std.call(here(), 'theme'));
      walk(here());
    end;
  }: with {
    obj {
      nam = '#f_less';
      act = function(s)
        prefs.font_scale = get_font_scale() - 10;
      
        prefs:store();
        set_theme(std.call(here(), 'theme'));
        walk(here());
      end;
    };
    obj {
      nam = '#f_more';
      act = function(s)
        prefs.font_scale = get_font_scale() + 10;
      
        prefs:store();
        set_theme(std.call(here(), 'theme'));
        walk(here());
      end;
    };
  };
  obj {
    nam = '#inv_font_scale';
    dsc = function(s)    
      local graphic_scale = get_graphic_scale();
      local font_scale = get_inventory_font_scale();
        
      pn(fmt.c(fmt.b(translate_now("menu_locale", 'inventory_scale')) .. ' {[' .. translate_now("menu_locale", 'auto_option') .. ']}'));
    
      local t = '';
    
      if font_scale > 50 then
        t = t .. '{#if_less|<<} ';
      else
        t = t .. '<< ';
      end;
    
      if prefs.inventory_scale then
        t = t .. tostring(font_scale) .. '%';
      else
        t = t .. translate_now("menu_locale", 'auto_option') .. ' (' .. tostring(font_scale) .. '%)';
      end;
    
    
      if font_scale < graphic_scale * 200 then
        t = t .. ' {#if_more|>>}';
      else
        t = t .. ' >>';
      end;
    
      pn(fmt.c(t) .. '^');
    end;
    act = function(s)
      prefs.inventory_scale = false;
    
      prefs:store();
      set_theme(std.call(here(), 'theme'));
      walk(here());
    end;
  }: with {
    obj {
      nam = '#if_less';
      act = function(s)
        prefs.inventory_scale = get_inventory_font_scale() - 10;
      
        prefs:store();
        set_theme(std.call(here(), 'theme'));
        walk(here());
      end;
    };
    obj {
      nam = '#if_more';
      act = function(s)
        prefs.inventory_scale = get_inventory_font_scale() + 10;
      
        prefs:store();
        set_theme(std.call(here(), 'theme'));
        walk(here());
      end;
    };
  };
  obj {
    nam = '#text_width';
    dsc = function(s)
      pn(fmt.c(fmt.b(translate_now("menu_locale", 'text_width'))));
      local t = '';
      if prefs.text_width > 40 then
        t = t .. '{#fw_less|<<} ';
      else
        t = t .. '<< ';
      end;

      t = t .. tostring(prefs.text_width) .. '%';
    
      if prefs.text_width < 100 then
        t = t .. ' {#fw_more|>>}';
      else
        t = t .. ' >>';
      end;
    
      pn(fmt.c(t) .. '^');
    end;
    act = function(s)
      if not prefs.text_width then
        prefs.text_width = 100;
      end;
    
      prefs.text_width = prefs.text_width + 20;
    
      if prefs.text_width > 100 then
        prefs.text_width = 40;
      end;
    
      prefs:store();
    
      walk(here());
    end;
  }: with {
    obj {
      nam = '#fw_less';
      act = function(s)
        prefs.text_width = prefs.text_width - 10;
      
        prefs:store();
        set_theme(std.call(here(), 'theme'));
        walk(here());
      end;
    };
    obj {
      nam = '#fw_more';
      act = function(s)
        prefs.text_width = prefs.text_width + 10;
      
        prefs:store();
        set_theme(std.call(here(), 'theme'));
        walk(here());
      end;
    };
  };
  obj {
    nam = '#text_alignment';
    dsc = function(s)
      pn(fmt.c(fmt.b(translate_now("menu_locale", 'text_alignment'))));
      
      local t = '';
    
      if prefs.text_alignment == 1 then
        t = t .. translate_now("menu_locale", 'text_alignment_justify') .. ' | ';
      else
        t = t .. '{#align_just|' .. translate_now("menu_locale", 'text_alignment_justify') .. '} | ';
      end;
      
    
      if prefs.text_alignment == 2 then
        t = t .. translate_now("menu_locale", 'text_alignment_left');
      else
        t = t .. '{#align_left|' .. translate_now("menu_locale", 'text_alignment_left') .. '}';
      end;
      pn(fmt.c(t) .. '^');
    end;
  }: with {
    obj {
      nam = '#align_just';
      act = function(s)
        prefs.text_alignment = 1;
      
        prefs:store();
        set_theme(std.call(here(), 'theme'));
        walk(here());
      end;
    };
    obj {
      nam = '#align_left';
      act = function(s)
        prefs.text_alignment = 2;
      
        prefs:store();
        set_theme(std.call(here(), 'theme'));
        walk(here());
      end;
    };
  };
  obj {
    nam = '#padding';
    dsc = function(s)
      pn(fmt.c(fmt.b(translate_now("menu_locale", 'padding'))));
      
      local t = '';
    
      if prefs.text_padding == 1 then
        t = t .. translate_now("menu_locale", 'normal') .. ' | ';
      else
        t = t .. '{#pad_norm|' .. translate_now("menu_locale", 'normal') .. '} | ';
      end;
      
    
      if prefs.text_padding == 2 then
        t = t .. translate_now("menu_locale", 'extended');
      else
        t = t .. '{#pad_large|' .. translate_now("menu_locale", 'extended') .. '}';
      end;
      pn(fmt.c(t) .. '^');
    end;
  }: with {
    obj {
      nam = '#pad_norm';
      act = function(s)
        prefs.text_padding = 1;
      
        prefs:store();
        set_theme(std.call(here(), 'theme'));
        walk(here());
      end;
    };
    obj {
      nam = '#pad_large';
      act = function(s)
        prefs.text_padding = 2;
      
        prefs:store();
        set_theme(std.call(here(), 'theme'));
        walk(here());
      end;
    };
  };
  obj {
    nam = '#interval';
    dsc = function(s)
      pn(fmt.c(fmt.b(translate_now("menu_locale", 'interval'))));
      
      local t = '';
    
      if prefs.text_interval == 1 then
        t = t .. translate_now("menu_locale", 'normal') .. ' | ';
      else
        t = t .. '{#int_norm|' .. translate_now("menu_locale", 'normal') .. '} | ';
      end;
      
    
      if prefs.text_interval == 2 then
        t = t .. translate_now("menu_locale", 'extended');
      else
        t = t .. '{#int_large|' .. translate_now("menu_locale", 'extended') .. '}';
      end;
      pn(fmt.c(t) .. '^');
    end;
  }: with {
    obj {
      nam = '#int_norm';
      act = function(s)
        prefs.text_interval = 1;
      
        prefs:store();
        set_theme(std.call(here(), 'theme'));
        walk(here());
      end;
    };
    obj {
      nam = '#int_large';
      act = function(s)
        prefs.text_interval = 2;
      
        prefs:store();
        set_theme(std.call(here(), 'theme'));
        walk(here());
      end;
    };
  };
  obj {
    nam = '#textinput';
    dsc = function(s)
      pn(fmt.c(fmt.b(translate_now("menu_locale", 'use_text_input'))));
      
      local t = '';
      if prefs.text_input then
        t = '{#ti_disable|' .. translate_now("menu_locale", 'disable') .. '} | ' .. translate_now("menu_locale", 'enable');
      else
        t = translate_now("menu_locale", 'disable') .. ' | {#ti_enable|' .. translate_now("menu_locale", 'enable') .. '}';
      end;
      
      pn(fmt.c(t) .. '^');
    end;
  }: with {
    obj {
      nam = '#ti_disable';
      act = function(s)
        prefs.text_input = false;
        prefs:store();
        walk(here());
      end;
    };
    obj {
      nam = '#ti_enable';
      act = function(s)
        prefs.text_input = true;
        prefs:store();
        walk(here());
      end;
    };
  };
  obj {
    nam = '#simplified_animation';
    dsc = function(s)
      pn(fmt.c(fmt.b(translate_now("menu_locale", 'simplified_animation'))));
      
      local t = '';
      if prefs.simplified_animation then
        t = '{#sa_disable|' .. translate_now("menu_locale", 'disable') .. '} | ' .. translate_now("menu_locale", 'enable');
      else
        t = translate_now("menu_locale", 'disable') .. ' | {#sa_enable|' .. translate_now("menu_locale", 'enable') .. '}';
      end;
      
      pn(fmt.c(t) .. '^');
    end;
  }: with {
    obj {
      nam = '#sa_disable';
      act = function(s)
        prefs.simplified_animation = false;
        prefs:store();
        walk(here());
      end;
    };
    obj {
      nam = '#sa_enable';
      act = function(s)
        prefs.simplified_animation = true;
        prefs:store();
        walk(here());
      end;
    };
  };
}
