include "terminal/base"

local input = std.ref '@input'
function input:text(char)
  return '@text "\\' .. char .. '"';
end;

std.mod_cmd(function(cmd)
  if cmd[1] == '@text' then
    if here().ontext then
      here():ontext(tostring(cmd[2]));
    end;
  end;
end)

function getscreen(graphical)
  local d = D('terminal_screen');
  if not d or d.graphical ~= graphical then
    local scale = diretheme.get_scale();
    local sw, sh = diretheme.get_scaled_screen();
    local x = math.floor((sw - 569) / 2 - 16);
    local dy = 0;
    if sh < 540 then
      dy = 19;
    end;
    local y = math.max(sh - 736 - dy, 25 - dy);
    local h = 320;
    if sh < 500 then
      h = h - 16;
    end;
    local offset = 0;
    if graphical then
      offset = 70 * scale;
    end;
    d = D {'terminal_screen', 'simple_txt', {}, lh=16 * scale, w = 569 * scale, h = h * scale, x = x * scale, y = y * scale, color = '#D0FFFF', color_link = '#abdece', color_alink = '#9bcebe', background=true, z=1000, font = 'fonts/vga.ttf', size=16 * scale, graphical = graphical, bgcolor='#000000', offset = offset};
  end;
  return d;
end;

function terminal_untheme(s)
  std.kh_leave_spaces = false;
  fmt.para = true;
end;

local function btn_anim_process(v)
  if v.animated == 1 then
    v.animated = true;
  elseif v.animated and v.frame_nr == 0 then
    v.animated = false;
    return true;
  else
    return true;
  end;
end;

local function shift_anim_process(v) 
  if v.frame_nr == 2 then
    v.animated = false;
  else
    return true;
  end;
end;

local function fake_anim_process(v)
  return true;
end;

local keymapping = {
  ['`'] = '~',
  ['1'] = '!',
  ['2'] = '@',
  ['3'] = '#',
  ['4'] = '$',
  ['5'] = '%',
  ['6'] = '^',
  ['7'] = '&',
  ['8'] = '*',
  ['9'] = '(',
  ['0'] = ')',
  ['-'] = '_',
  ['='] = '+',
  ['q'] = 'Q', 
  ['w'] = 'W', 
  ['e'] = 'E',
  ['r'] = 'R',
  ['t'] = 'T',
  ['y'] = 'Y',
  ['u'] = 'U',
  ['i'] = 'I', 
  ['o'] = 'O',
  ['p'] = 'P',
  ['['] = '{',
  [']'] = '}',
  ['\\'] = '|',
  ['a'] = 'A',
  ['s'] = 'S',
  ['d'] = 'D',
  ['f'] = 'F',
  ['g'] = 'G',
  ['h'] = 'H',
  ['j'] = 'J',
  ['k'] = 'K',
  ['l'] = 'L',
  [';'] = ':',
  ["'"] = '"',
  ['z'] = 'Z',
  ['x'] = 'X',
  ['c'] = 'C',
  ['v'] = 'V',
  ['b'] = 'B',
  ['n'] = 'N',
  ['m'] = 'M',
  [','] = '.',
  ['.'] = ',',
  ['/'] = '?',
};

local reverse_keymapping = {
  ['>'] = '>';
  ['<'] = '<'
};
for k, v in pairs(keymapping) do
  reverse_keymapping[v] = k;
end;

local key_sounds = {
  'keyboard1', 'keyboard2', 'keyboard3', 'keyboard4', 'keyboard5'
}

local handled_keys = Set {
  '`', '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', '-', '=', 'backspace','shift',
  'q', 'w', 'e', 'r', 't', 'y', 'u','i', 'o', 'p', '[', ']', '\\', 
  'a', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', ';', "'", 'return',
  'z', 'x', 'c', 'v', 'b', 'n', 'm', ',', '.', '/', 
  'space', 'left', 'right', 'up', 'down', 'left shift'
}

local release_handled_keys = Set {
  'left shift'
}

local mapping_decor_key = {
  btn_backquote = '`';
  btn_1 = '1';
  btn_2 = '2';
  btn_3 = '3';
  btn_4 = '4';
  btn_5 = '5';
  btn_6 = '6';
  btn_7 = '7';
  btn_8 = '8';
  btn_9 = '9';
  btn_0 = '0';
  btn_minus = '-';
  btn_backspace = 'backspace';
  
  btn_q = 'q';
  btn_w = 'w';
  btn_e = 'e';
  btn_r = 'r';
  btn_t = 't';
  btn_y = 'y';
  btn_u = 'u';
  btn_i = 'i';
  btn_o = 'o';
  btn_p = 'p';
  btn_left_bracket = '[';
  btn_right_bracket = ']';
  btn_equal = '=';

  btn_a = 'a';
  btn_s = 's';
  btn_d = 'd';
  btn_f = 'f';
  btn_g = 'g';
  btn_h = 'h';
  btn_j = 'j';
  btn_k = 'k';
  btn_l = 'l';
  btn_semicolon = ';';
  btn_enter = 'return';


  btn_z = 'z';
  btn_x = 'x';
  btn_c = 'c';
  btn_v = 'v';
  btn_b = 'b';
  btn_n = 'n';
  btn_m = 'm';
  btn_comma = ',';
  btn_slash = '/';
  btn_backslash = '\\';
  btn_quote = '\'';
  btn_space = 'space';
};

local back_mapping_decor_key = {};

for k, v in pairs(mapping_decor_key) do
  back_mapping_decor_key[v] = k;
end;

function click:filter(press, btn, x, y, px, py)
  return not press
end

room {
  {
    ticks = 0;
    ram = 0;
  };
  nam = 'terminal_boot';
  disp = '';
  nosave = true;
  noautosave = true;
  nofollow = true;
  noinv = true;
  nofading = true;
  network = false;
  graphical = false;
  theme = 'terminal';
  themeInit = function(s)
    load_sounds('error_beep', 'pc_startup', 'keyboard1', 'keyboard2', 'keyboard3', 'keyboard4', 'keyboard5');
    timer:set(25);
    
    local scale = diretheme.get_scale();
    local width, height = diretheme.get_scaled_screen();
    
    local rc = 0;
    if s.graphical then
      local animx = math.max(math.floor((width - 640) / 2) + 20, 20) * scale;
      local animy = math.max(height - 734, 26) * scale;
      
      if height < 540 then
        animy = animy - 19 * scale;
      end;
      D {'logo', 'img', get_sprite('images/terminal/pc_anim.png'), x = animx, y = animy, w=64 * scale, h=64 * scale, frames = 24, z = 1, delay=175, speed=175, animated=true, background=true, process=function(d)
        if d.frame_nr == 8 and rc < 2 then
          d.frame_nr = 0;
          rc = rc + 1;
        elseif d.frame_nr == 9 then
          rc = 0;
        end;
      end};
    end;
    
    local x = math.floor((width - 800) / 2);
    
    local y = height - 193;
    if y < 564 then
      y = 564;
    end;
    
    local top = height - 193;
    if height < 540 then
      top = top + 20;
    end;

    local delta = math.ceil((y - top) / 8) * scale;
    top = top * scale;
    
    DS {'keyboard', 'img', 'images/terminal/keyboard.png', x = x - 116, y = y, w=1036, h=191, z=0, speed=25, process=function(d)
      if s.network and s.ticks >= 128 and delta > 0 then
        d.y = d.y - delta;
        delta = delta - scale;
        if d.y < top then
           d.y = top;
           delta = 0;
        end;
      else
        return true;
      end;
    end};
  end;
  onenter = function(s)
    s.ticks = 0;
    s.ram = 0;
    local d = getscreen(s.graphical);
    d:drop_all();
    
    stop_sound();
  end;
  onexit = function(s)
    D {'logo'};
    D {'keyboard'};
    D {'terminal_screen'};
  end;
  timer = function(s)
    if s.ticks ~= 33 then
      s.ticks = s.ticks + 1;
    else
      local ram = lookup('#ram', _'system_unit').device;

      s.ram = s.ram + ram.size * 16;
      if s.ram > ram.size * 1024 then
        s.ram = ram.size * 1024;
        s.ticks = s.ticks + 1;
      end;
    end;
    if s.network then
      if s.ticks == 128 then
        D {'logo'};
      end;
      if s.ticks >= 156 then
        walk('terminal');
      end;
    else
      if s.ticks >= 300 then
        walk('tower_level_4_room_2');
        return;
      end;
      if s.ticks > 112 then
        return false;
      end;
    end;
    
    s:makescreen();
    
    return false;
  end;
  dsc = '';
  txt = {};
  makescreen = function(s)
    local d = getscreen(s.graphical and s.ticks < 128);
    if s.ticks < 128 then
      local motherboard = lookup('#motherboard', _'system_unit').device;
      local cpu = lookup('#cpu', _'system_unit').device;
      local ram = lookup('#ram', _'system_unit').device;
      
      if s.ticks == 1 then
        d:insert_line('TOWER BIOS 1.0');
      end;

      if s.ticks == 8 then
        d:insert_line(translate_now("terminal_locale", 'motherboard') .. ' : ' .. motherboard.name);
      end;

      if s.ticks == 16 then
        d:insert_line(translate_now("terminal_locale", 'cpu') .. ' : ' .. cpu.name .. ' @ ' .. cpu.speed .. '.0' .. translate_now("terminal_locale", 'mhz'));
      end;

      if s.ticks == 24 then
        d:insert_line(translate_now("terminal_locale", 'ram') .. ' : '.. ram.name);
      end;

      if s.ticks == 32 then
        d:insert_line(tostring(s.ram) .. translate_now("terminal_locale", 'kilobytes'))
      end;

      if s.ticks == 33 then
        d:replace_line(tostring(s.ram) .. translate_now("terminal_locale", 'kilobytes'))
      end;

      if s.ticks == 34 then
        d:replace_line(tostring(s.ram) .. translate_now("terminal_locale", 'kilobytes'));
      end;

      if s.ticks == 40 then
        d:replace_line(tostring(s.ram) .. translate_now("terminal_locale", 'kilobytes') .. ' OK');
        d:insert_line(' ');
      end;

      if s.ticks == 48 then
        d:insert_line(translate_now("terminal_locale", 'keyboard') .. ' : ');
      end;

      if s.ticks == 56 then
        d:replace_line(translate_now("terminal_locale", 'keyboard') .. ' : OK')
      end;

      if s.ticks == 64 then
        d:insert_line(translate_now("terminal_locale", 'disk') .. ' : ');
      end;

      if s.ticks == 72 then
        d:replace_line(translate_now("terminal_locale", 'disk') .. ' : ' .. translate_now("terminal_locale", 'missing'));
      end;

      if s.ticks == 80 then
        d:insert_line(translate_now("terminal_locale", 'network') .. ' : ');
      end;

      if s.ticks == 88 then
        if s.network then
          d:replace_line(translate_now("terminal_locale", 'network') .. ' : OK');
        else
          d:replace_line(translate_now("terminal_locale", 'network') .. ' : ' .. translate_now("terminal_locale", 'missing'));
        end;
        d:insert_line(' ');
      end;

      if s.ticks == 104 then
        d:insert_line(translate_now("terminal_locale", 'boot_from'));
      end;

      if s.network then
        if s.ticks == 112 then
          d:replace_line(translate_now("terminal_locale", 'boot_from') .. ' ' .. translate_now("terminal_locale", 'network_2'));
        end;

        if s.ticks == 120 then
          d:replace_line(translate_now("terminal_locale", 'boot_from') .. ' ' .. translate_now("terminal_locale", 'network_2') .. '...');
        end;
      else
        if s.ticks == 112 then
          d:replace_line(translate_now("terminal_locale", 'boot_from') .. ' ' .. translate_now("terminal_locale", 'no_device'));
        end;
      end;
    elseif s.ticks == 128 then
      d:drop_all();
      d:insert_line(translate_now("terminal_locale", 'tower_os_line_1'));
      d:insert_line(' ');
    else
      if s.ticks == 136 then
        d:insert_line(translate_now("terminal_locale", 'tower_os_line_2'));
        d:insert_line(' ');
      end;

      if s.ticks == 144 then
        d:insert_line(translate_now("terminal_locale", 'tower_os_line_3'));
        d:insert_line(' ');
      end;
    end;
  end;
}

room {
  {
    ticks = 0;
    ram = 0;
  };
  nam = 'terminal_boot_failure';
  theme = 'terminal';
  disp = '';
  noinv = true;
  nofading = true;
  nofollow = true;
  graphical = false;
  nosave = true;
  noautosave = true;
  themeInit = function(s)
    timer:set(25);
    
    local scale = diretheme.get_scale();
    local width, height = diretheme.get_scaled_screen();
    
    local rc = 0;
    if s.graphical then
      local animx = math.max(math.floor((width - 640) / 2) + 20, 20) * scale;
      local animy = math.max(height - 734, 26) * scale;
      
      if height < 540 then
        animy = animy - 19 * scale;
      end;
      D {'logo', 'img', get_sprite('images/terminal/pc_anim.png'), x = animx, y = animy, w=64 * scale, h=64 * scale, frames = 24, z = 1, delay = 175, speed = 175, animated=true, background=true, process=function(d)
        if d.frame_nr == 8 and rc < 2 then
          d.frame_nr = 0;
          rc = rc + 1;
        elseif d.frame_nr == 9 then
          rc = 0;
        end;
      end};
    end;

    local x = math.floor((width - 800) / 2);
    
    local y = height - 193;
    if height < 540 then
      y = y + 20;
    end;
    
    DS {'keyboard', 'img', 'images/terminal/keyboard.png', x = x - 116, y = y, w=1036, h=191, z=0};
  end;
  onenter = function(s)
    s.graphical = _('terminal_boot').graphical;
    s.ticks = 0;
    s.ram = 0;
    local d = getscreen(s.graphical);
    d:drop_all();
    
    stop_sound();
  end;
  onexit = function(s)    
    enableSmoke('images/endings/tower_ruin.png');

    D {'terminal_screen'};
    D {'keyboard'};
    D {'logo'};
  end;
  timer = function(s)
    if s.ticks ~= 33 then
      s.ticks = s.ticks + 1;
    else
      local ram = lookup('#ram', _'system_unit').device;

      s.ram = s.ram + ram.size * 16;
      if s.ram > ram.size * 1024 then
        s.ram = ram.size * 1024;
        s.ticks = s.ticks + 1;
      end;
    end;

    if s.ticks == 128 then
      D {'logo'};
    end;
    if s.ticks >= 200 then
      walk('ending_formatted');
      return;
    end;

    s:makescreen();
    
    return false;
  end;
  dsc = '';
  txt = '';
  makescreen = function(s)
    local d = getscreen(s.graphical and s.ticks < 128);
    if s.ticks < 128 then
      local motherboard = lookup('#motherboard', _'system_unit').device;
      local cpu = lookup('#cpu', _'system_unit').device;
      local ram = lookup('#ram', _'system_unit').device;

      if s.ticks == 1 then
        d:insert_line('TOWER BIOS 1.0');
      end;

      if s.ticks == 8 then
        d:insert_line(translate_now("terminal_locale", 'motherboard') .. ' : ' .. motherboard.name);
      end;

      if s.ticks == 16 then
        d:insert_line(translate_now("terminal_locale", 'cpu') .. ' : ' .. cpu.name .. ' @ ' .. cpu.speed .. '.0' .. translate_now("terminal_locale", 'mhz'));
      end;

      if s.ticks == 24 then
        d:insert_line(translate_now("terminal_locale", 'ram') .. ' : '.. ram.name);
      end;

      if s.ticks == 32 then
        d:insert_line(tostring(s.ram) .. translate_now("terminal_locale", 'kilobytes'))
      end;

      if s.ticks == 33 then
        d:replace_line(tostring(s.ram) .. translate_now("terminal_locale", 'kilobytes'))
      end;

      if s.ticks == 34 then
        d:replace_line(tostring(s.ram) .. translate_now("terminal_locale", 'kilobytes'));
      end;

      if s.ticks == 40 then
        d:replace_line(tostring(s.ram) .. translate_now("terminal_locale", 'kilobytes') .. ' OK');
        d:insert_line(' ');
      end;

      if s.ticks == 48 then
        d:insert_line(translate_now("terminal_locale", 'keyboard') .. ' : ');
      end;

if s.ticks == 56 then
        d:replace_line(translate_now("terminal_locale", 'keyboard') .. ' : OK')
      end;

      if s.ticks == 64 then
        d:insert_line(translate_now("terminal_locale", 'disk') .. ' : ');
      end;

      if s.ticks == 72 then
        d:replace_line(translate_now("terminal_locale", 'disk') .. ' : ' .. translate_now("terminal_locale", 'missing'));
      end;

      if s.ticks == 80 then
        d:insert_line(translate_now("terminal_locale", 'network') .. ' : ');
      end;

      if s.ticks == 88 then
        if _('terminal_boot').network then
          d:replace_line(translate_now("terminal_locale", 'network') .. ' : OK');
        else
          d:replace_line(translate_now("terminal_locale", 'network') .. ' : ' .. translate_now("terminal_locale", 'missing'));
        end;
        d:insert_line(' ');
      end;

      if s.ticks == 104 then
        d:insert_line(translate_now("terminal_locale", 'boot_from'));
      end;


      if s.ticks == 112 then
        d:replace_line(translate_now("terminal_locale", 'boot_from') .. ' ' .. translate_now("terminal_locale", 'network_2'));
      end;

      if s.ticks == 120 then
        d:replace_line(translate_now("terminal_locale", 'boot_from') .. ' ' .. translate_now("terminal_locale", 'network_2') .. '...');
      end;
    else
      d:drop_all();
      d:insert_line(translate_now("terminal_locale", 'no_bootable_medium_found'));
    end;
  end;
}

terminal {
  nam = 'terminal';
  
  keysfilter = function(s, press, key)
    return handled_keys[key] ~= nil; -- ловим все нажатия
  end;
  
  timer = function(s)
    if s.status == 'wait_format' then
      if s.format_percent < 100 then
        s.format_percent = s.format_percent + 1;
        local d = getscreen(false);
        d:replace_line(s:query_formatter(s.path, s.entry));
      else
        s.status = 'format_done';
        s:push(translate_now("terminal_locale", 'formatting_complete'));
      end;
      return false;
    elseif s.status == 'format_done' then
      if s.format_percent < 120 then
        s:push(translate_now("terminal_locale", 'critical_failure'));
      end;
      
      s.format_percent = s.format_percent + 1;
      if s.format_percent > 150 then
        timer:stop();
        walk('terminal_boot_failure');
        return;
      end;
      return false;
    end;
    return false;
  end;
  
  push_line = function(s, line)
    local d = getscreen(false);
    d:replace_line(line);
    if s.query_formatter then
      d:insert_line(s:query_formatter(s.path, s.entry));
    else
      d:insert_line(s.path .. '>' .. s.entry .. '_');
    end;
  end;
  
  makekeyboard = function(s)
    local width, height = diretheme.get_scaled_screen();
    
    local x = math.floor((width - 800) / 2);
    local y = height - 191;
    if height < 540 then
      y = y + 20;
    end;
    
    local process = btn_anim_process;
    local shift_process = shift_anim_process;
    local decortype = 'img';
    if prefs.simplified_animation then
      if keys:state('shift') then
        DS {'keyboard_bg', 'img', 'images/terminal/keyboard_shift.png', x=x - 116, y=y - 2, w=1036, h=191, z=110};
      else
        DS {'keyboard_bg', 'img', 'images/terminal/keyboard.png', x=x - 116, y=y - 2, w=1036, h=191, z=110};
      end;
      decortype = 'raw';
      
      process = fake_anim_process;
      shift_process = fake_anim_process;
    else
      DS {'keyboard_bg', 'img', 'images/terminal/keyboard_bg.png', x=x - 116, y=y - 2, w=1036, h=191, z=110};
      DS {'panel_1', 'img', 'images/terminal/keyboard/panel_1.png', x=x + 62, y=y + 37, w=677, h=16, z=92};
      DS {'panel_2', 'img', 'images/terminal/keyboard/panel_2.png', x=x + 77, y=y + 62, w=648, h=12,  z=81};
      DS {'panel_3', 'img', 'images/terminal/keyboard/panel_3.png', x=x + 78, y=y + 94, w=652, h=20, z=72};
      DS {'panel_4', 'img', 'images/terminal/keyboard/panel_4.png', x=x + 78, y=y + 131, w=652, h=23, z=61};
      DS {'panel_5', 'img', 'images/terminal/keyboard/panel_5.png', x=x, y=y + 181, w=800, h=10, z=51};
    end;
    
    -- Keyboard row 1
    -- D {'btn_backquote', 'img', 'images/terminal/keyboard/btn_backquote.png', x=x + 67, y=y, w=64, h=46, frames=3, z=100, delay=50, process=process, click=true};
    DS {'btn_1', decortype, 'images/terminal/keyboard/btn_1.png', x=x + 119, y=y, w=58, h=46, frames=3, z=99, delay=50, process=process, click=true};
    DS {'btn_2', decortype, 'images/terminal/keyboard/btn_2.png', x=x + 167, y=y, w=55, h=46, frames=3, z=98, delay=50, process=process, click=true};
    DS {'btn_3', decortype, 'images/terminal/keyboard/btn_3.png', x=x + 215, y=y, w=51, h=46, frames=3, z=97, delay=50, process=process, click=true};
    DS {'btn_4', decortype, 'images/terminal/keyboard/btn_4.png', x=x + 260, y=y, w=49, h=46, frames=3, z=96, delay=50, process=process, click=true};
    DS {'btn_5', decortype, 'images/terminal/keyboard/btn_5.png', x=x + 307, y=y, w=44, h=46, frames=3, z=95, delay=50, process=process, click=true};
    DS {'btn_6', decortype, 'images/terminal/keyboard/btn_6.png', x=x + 351, y=y, w=47, h=46, frames=3, z=95, delay=50, process=process, click=true};
    DS {'btn_7', decortype, 'images/terminal/keyboard/btn_7.png', x=x + 398, y=y, w=48, h=46, frames=3, z=96, delay=50, process=process, click=true};
    DS {'btn_8', decortype, 'images/terminal/keyboard/btn_8.png', x=x + 444, y=y, w=48, h=46, frames=3, z=97, delay=50, process=process, click=true};
    DS {'btn_9', decortype, 'images/terminal/keyboard/btn_9.png', x=x + 486, y=y, w=51, h=46, frames=3, z=98, delay=50, process=process, click=true};
    DS {'btn_0', decortype, 'images/terminal/keyboard/btn_0.png', x=x + 529, y=y, w=55, h=46, frames=3, z=99, delay=50, process=process, click=true};
    -- DS {'btn_minus', decortype, 'images/terminal/keyboard/btn_minus.png', x=x + 574, y=y, w=58, h=46, frames=3, z=100, delay=50, process=process, click=true};
    DS {'btn_backspace', decortype, 'images/terminal/keyboard/btn_backspace.png', x=x + 619, y=y, w=109, h=46, frames=3, z=101, delay=50, process=process, click=true};

    -- Keyboard row 2
    DS {'btn_q', decortype, 'images/terminal/keyboard/btn_q.png', x=x + 81, y=y + 26, w=61, h=46, frames=3, z=90, delay=50, process=process, click=true};
    DS {'btn_w', decortype, 'images/terminal/keyboard/btn_w.png', x=x + 130, y=y + 26, w=59, h=46, frames=3, z=89, delay=50, process=process, click=true};
    DS {'btn_e', decortype, 'images/terminal/keyboard/btn_e.png', x=x + 180, y=y + 26, w=57, h=46, frames=3, z=88, delay=50, process=process, click=true};
    DS {'btn_r', decortype, 'images/terminal/keyboard/btn_r.png', x=x + 230, y=y + 26, w=53, h=46, frames=3, z=87, delay=50, process=process, click=true};
    DS {'btn_t', decortype, 'images/terminal/keyboard/btn_t.png', x=x + 278, y=y + 26, w=50, h=46, frames=3, z=86, delay=50, process=process, click=true};
    DS {'btn_y', decortype, 'images/terminal/keyboard/btn_y.png', x=x + 327, y=y + 26, w=45, h=46, frames=3, z=85, delay=50, process=process, click=true};
    DS {'btn_u', decortype, 'images/terminal/keyboard/btn_u.png', x=x + 373, y=y + 26, w=52, h=46, frames=3, z=85, delay=50, process=process, click=true};
    DS {'btn_i', decortype, 'images/terminal/keyboard/btn_i.png', x=x + 423, y=y + 26, w=52, h=46, frames=3, z=86, delay=50, process=process, click=true};
    DS {'btn_o', decortype, 'images/terminal/keyboard/btn_o.png', x=x + 471, y=y + 26, w=50, h=46, frames=3, z=87, delay=50, process=process, click=true};
    DS {'btn_p', decortype, 'images/terminal/keyboard/btn_p.png', x=x + 514, y=y + 26, w=58, h=46, frames=3, z=88, delay=50, process=process, click=true};
    -- DS {'btn_left_bracket', decortype, 'images/terminal/keyboard/btn_left_bracket.png', x=x + 562, y=y + 26, w=60, h=46, frames=3, z=89, delay=50, process=process, click=true};
    -- DS {'btn_right_bracket', decortype, 'images/terminal/keyboard/btn_right_bracket.png', x=x + 610, y=y + 26, w=62, h=46, frames=3, z=90, delay=50, process=process, click=true};
    -- DS {'btn_equal', decortype, 'images/terminal/keyboard/btn_equal.png', x=x + 656, y=y + 26, w=65, h=46, frames=3, z=91, delay=50, process=process, click=true};
    
    -- Keyboard row 3
    DS {'btn_a', decortype, 'images/terminal/keyboard/btn_a.png', x=x + 81, y=y + 52, w=69, h=55, frames=3, z=80, delay=50, process=process, click=true};
    DS {'btn_s', decortype, 'images/terminal/keyboard/btn_s.png', x=x + 135, y=y + 52, w=67, h=55, frames=3, z=79, delay=50, process=process, click=true};
    DS {'btn_d', decortype, 'images/terminal/keyboard/btn_d.png', x=x + 191, y=y + 52, w=61, h=55, frames=3, z=78, delay=50, process=process, click=true};
    DS {'btn_f', decortype, 'images/terminal/keyboard/btn_f.png', x=x + 244, y=y + 52, w=58, h=55, frames=3, z=77, delay=50, process=process, click=true};
    DS {'btn_g', decortype, 'images/terminal/keyboard/btn_g.png', x=x + 296, y=y + 52, w=53, h=55, frames=3, z=76, delay=50, process=process, click=true};
    DS {'btn_h', decortype, 'images/terminal/keyboard/btn_h.png', x=x + 346, y=y + 52, w=55, h=55, frames=3, z=75, delay=50, process=process, click=true};
    DS {'btn_j', decortype, 'images/terminal/keyboard/btn_j.png', x=x + 401, y=y + 52, w=56, h=55, frames=3, z=76, delay=50, process=process, click=true};
    DS {'btn_k', decortype, 'images/terminal/keyboard/btn_k.png', x=x + 452, y=y + 52, w=55, h=55, frames=3, z=77, delay=50, process=process, click=true};
    DS {'btn_l', decortype, 'images/terminal/keyboard/btn_l.png', x=x + 500, y=y + 52, w=60, h=55, frames=3, z=78, delay=50, process=process, click=true};
    -- DS {'btn_semicolon', decortype, 'images/terminal/keyboard/btn_semicolon.png', x=x + 549, y=y + 52, w=65, h=55, frames=3, z=79, delay=50, process=process, click=true};
    DS {'btn_enter', decortype, 'images/terminal/keyboard/btn_enter.png', x=x + 599, y=y + 52, w=124, h=55, frames=3, z=80, delay=50, process=process, click=true};

    -- Keyboard row 4
    DS {'btn_z', decortype, 'images/terminal/keyboard/btn_z.png', x=x + 89, y=y + 86, w=74, h=59, frames=3, z=70, delay=50, process=process, click=true};
    DS {'btn_x', decortype, 'images/terminal/keyboard/btn_x.png', x=x + 147, y=y + 86, w=70, h=59, frames=3, z=69, delay=50, process=process, click=true};
    DS {'btn_c', decortype, 'images/terminal/keyboard/btn_c.png', x=x + 205, y=y + 86, w=64, h=59, frames=3, z=68, delay=50, process=process, click=true};
    DS {'btn_v', decortype, 'images/terminal/keyboard/btn_v.png', x=x + 261, y=y + 86, w=61, h=59, frames=3, z=67, delay=50, process=process, click=true};
    DS {'btn_b', decortype, 'images/terminal/keyboard/btn_b.png', x=x + 318, y=y + 86, w=53, h=59, frames=3, z=66, delay=50, process=process, click=true};
    DS {'btn_n', decortype, 'images/terminal/keyboard/btn_n.png', x=x + 372, y=y + 86, w=60, h=59, frames=3, z=66, delay=50, process=process, click=true};
    DS {'btn_m', decortype, 'images/terminal/keyboard/btn_m.png', x=x + 429, y=y + 86, w=61, h=59, frames=3, z=67, delay=50, process=process, click=true};
    -- DS {'btn_comma', decortype, 'images/terminal/keyboard/btn_comma.png', x=x + 483, y=y + 86, w=62, h=59, frames=3, z=68, delay=50, process=process, click=true};
    -- DS {'btn_slash', decortype, 'images/terminal/keyboard/btn_slash.png', x=x + 534, y=y + 86, w=69, h=59, frames=3, z=69, delay=50, process=process, click=true};
    -- DS {'btn_backslash', decortype, 'images/terminal/keyboard/btn_backslash.png', x=x + 589, y=y + 86, w=74, h=59, frames=3, z=70, delay=50, process=process, click=true};
    DS {'btn_quote', decortype, 'images/terminal/keyboard/btn_quote.png', x=x + 643, y=y + 86, w=78, h=59, frames=3, z=71, delay=50, process=process, click=true};

    local shift_frame = 0;
    if keys:state('shift') then
      shift_frame = 2;
    end;

    DS {'btn_shift', decortype, 'images/terminal/keyboard/btn_shift.png', x=x + 94, y=y + 125, w=139, h=64, frames=3, z=60, delay=50, process=shift_process, click=true, frame_nr=shift_frame};
    DS {'btn_space', decortype, 'images/terminal/keyboard/btn_space.png', x=x + 222, y=y + 125, w=494, h=64, frames=3, z=59, delay=50, process=process, click=true};

    if keys:state('shift') then
      DS {'btn_backquote', decortype, 'images/terminal/keyboard/btn_tilde.png', x=x + 67, y=y, w=64, h=46, frames=3, z=100, delay=50, process=process, click=true};
      DS {'btn_minus', decortype, 'images/terminal/keyboard/btn_underscore.png', x=x + 574, y=y, w=58, h=46, frames=3, z=100, delay=50, process=process, click=true};
      DS {'btn_left_bracket', decortype, 'images/terminal/keyboard/btn_left_curly.png', x=x + 562, y=y + 26, w=60, h=46, frames=3, z=89, delay=50, process=process, click=true};
      DS {'btn_right_bracket', decortype, 'images/terminal/keyboard/btn_right_curly.png', x=x + 610, y=y + 26, w=62, h=46, frames=3, z=90, delay=50, process=process, click=true};
      DS {'btn_equal', decortype, 'images/terminal/keyboard/btn_plus.png', x=x + 656, y=y + 26, w=65, h=46, frames=3, z=91, delay=50, process=process, click=true};
      DS {'btn_semicolon', decortype, 'images/terminal/keyboard/btn_colon.png', x=x + 549, y=y + 52, w=65, h=55, frames=3, z=79, delay=50, process=process, click=true};
      DS {'btn_comma', decortype, 'images/terminal/keyboard/btn_dot.png', x=x + 483, y=y + 86, w=62, h=59, frames=3, z=68, delay=50, process=process, click=true};
      DS {'btn_slash', decortype, 'images/terminal/keyboard/btn_question.png', x=x + 534, y=y + 86, w=69, h=59, frames=3, z=69, delay=50, process=process, click=true};
      DS {'btn_backslash', decortype, 'images/terminal/keyboard/btn_vertical.png', x=x + 589, y=y + 86, w=74, h=59, frames=3, z=70, delay=50, process=process, click=true};
    else
      DS {'btn_backquote', decortype, 'images/terminal/keyboard/btn_backquote.png', x=x + 67, y=y, w=64, h=46, frames=3, z=100, delay=50, process=process, click=true};
      DS {'btn_minus', decortype, 'images/terminal/keyboard/btn_minus.png', x=x + 574, y=y, w=58, h=46, frames=3, z=100, delay=50, process=process, click=true};
      DS {'btn_left_bracket', decortype, 'images/terminal/keyboard/btn_left_bracket.png', x=x + 562, y=y + 26, w=60, h=46, frames=3, z=89, delay=50, process=process, click=true};
      DS {'btn_right_bracket', decortype, 'images/terminal/keyboard/btn_right_bracket.png', x=x + 610, y=y + 26, w=62, h=46, frames=3, z=90, delay=50, process=process, click=true};
      DS {'btn_equal', decortype, 'images/terminal/keyboard/btn_equal.png', x=x + 656, y=y + 26, w=65, h=46, frames=3, z=91, delay=50, process=process, click=true};
      DS {'btn_semicolon', decortype, 'images/terminal/keyboard/btn_semicolon.png', x=x + 549, y=y + 52, w=65, h=55, frames=3, z=79, delay=50, process=process, click=true};
      DS {'btn_comma', decortype, 'images/terminal/keyboard/btn_comma.png', x=x + 483, y=y + 86, w=62, h=59, frames=3, z=68, delay=50, process=process, click=true};
      DS {'btn_slash', decortype, 'images/terminal/keyboard/btn_slash.png', x=x + 534, y=y + 86, w=69, h=59, frames=3, z=69, delay=50, process=process, click=true};
      DS {'btn_backslash', decortype, 'images/terminal/keyboard/btn_backslash.png', x=x + 589, y=y + 86, w=74, h=59, frames=3, z=70, delay=50, process=process, click=true};
    end;
  end;
  onexit = function(s)
    D {'keyboard_bg'};
    -- Keyboard row 1
    D {'btn_backquote'};
    D {'btn_1'};
    D {'btn_2'};
    D {'btn_3'};
    D {'btn_4'};
    D {'btn_5'};
    D {'btn_6'};
    D {'btn_7'};
    D {'btn_8'};
    D {'btn_9'};
    D {'btn_0'};
    D {'btn_minus'};
    D {'btn_backspace'};

    -- Keyboard row 2
    D {'btn_q'};
    D {'btn_w'};
    D {'btn_e'};
    D {'btn_r'};
    D {'btn_t'};
    D {'btn_y'};
    D {'btn_u'};
    D {'btn_i'};
    D {'btn_o'};
    D {'btn_p'};
    D {'btn_left_bracket'};
    D {'btn_right_bracket'};
    D {'btn_equal'};

    D {'btn_a'};
    D {'btn_s'};
    D {'btn_d'};
    D {'btn_f'};
    D {'btn_g'};
    D {'btn_h'};
    D {'btn_j'};
    D {'btn_k'};
    D {'btn_l'};
    D {'btn_semicolon'};
    D {'btn_enter'};

    D {'btn_z'};
    D {'btn_x'};
    D {'btn_c'};
    D {'btn_v'};
    D {'btn_b'};
    D {'btn_n'};
    D {'btn_m'};
    D {'btn_comma'};
    D {'btn_slash'};
    D {'btn_backslash'};
    D {'btn_quote'};

    D {'btn_shift'};
    D {'btn_space'};

    D {'panel_1'};
    D {'panel_2'};
    D {'panel_3'};
    D {'panel_4'};
    D {'panel_5'};
    D {'terminal_screen'};
  end;
  shift_buttons = function(s)
    local width, height = diretheme.get_scaled_screen();
    
    local x = math.floor((width - 800) / 2);
    local y = height - 191;
    if height < 540 then
      y = y + 20;
    end;

    if prefs.simplified_animation then
      if keys:state('shift') then
        DS {'keyboard_bg', 'img', 'images/terminal/keyboard_shift.png', x=x - 116, y=y - 2, w=1036, h=191, z=110};
      else
        DS {'keyboard_bg', 'img', 'images/terminal/keyboard.png', x=x - 116, y=y - 2, w=1036, h=191, z=110};
      end;
      return;
    end;

    if keys:state('shift') then
      DS {'btn_backquote', 'img', 'images/terminal/keyboard/btn_tilde.png', x=x + 67, y=y, w=64, h=46, frames=3, z=100, delay=50, process=btn_anim_process, click=true};
      DS {'btn_minus', 'img', 'images/terminal/keyboard/btn_underscore.png', x=x + 574, y=y, w=58, h=46, frames=3, z=100, delay=50, process=btn_anim_process, click=true};
      DS {'btn_left_bracket', 'img', 'images/terminal/keyboard/btn_left_curly.png', x=x + 562, y=y + 26, w=60, h=46, frames=3, z=89, delay=50, process=btn_anim_process, click=true};
      DS {'btn_right_bracket', 'img', 'images/terminal/keyboard/btn_right_curly.png', x=x + 610, y=y + 26, w=62, h=46, frames=3, z=90, delay=50, process=btn_anim_process, click=true};
      DS {'btn_equal', 'img', 'images/terminal/keyboard/btn_plus.png', x=x + 656, y=y + 26, w=65, h=46, frames=3, z=91, delay=50, process=btn_anim_process, click=true};
      DS {'btn_semicolon', 'img', 'images/terminal/keyboard/btn_colon.png', x=x + 549, y=y + 52, w=65, h=55, frames=3, z=79, delay=50, process=btn_anim_process, click=true};
      DS {'btn_comma', 'img', 'images/terminal/keyboard/btn_dot.png', x=x + 483, y=y + 86, w=62, h=59, frames=3, z=68, delay=50, process=btn_anim_process, click=true};
      DS {'btn_slash', 'img', 'images/terminal/keyboard/btn_question.png', x=x + 534, y=y + 86, w=69, h=59, frames=3, z=69, delay=50, process=btn_anim_process, click=true};
      DS {'btn_backslash', 'img', 'images/terminal/keyboard/btn_vertical.png', x=x + 589, y=y + 86, w=74, h=59, frames=3, z=70, delay=50, process=btn_anim_process, click=true};
    else
      DS {'btn_backquote', 'img', 'images/terminal/keyboard/btn_backquote.png', x=x + 67, y=y, w=64, h=46, frames=3, z=100, delay=50, process=btn_anim_process, click=true};
      DS {'btn_minus', 'img', 'images/terminal/keyboard/btn_minus.png', x=x + 574, y=y, w=58, h=46, frames=3, z=100, delay=50, process=btn_anim_process, click=true};
      DS {'btn_left_bracket', 'img', 'images/terminal/keyboard/btn_left_bracket.png', x=x + 562, y=y + 26, w=60, h=46, frames=3, z=89, delay=50, process=btn_anim_process, click=true};
      DS {'btn_right_bracket', 'img', 'images/terminal/keyboard/btn_right_bracket.png', x=x + 610, y=y + 26, w=62, h=46, frames=3, z=90, delay=50, process=btn_anim_process, click=true};
      DS {'btn_equal', 'img', 'images/terminal/keyboard/btn_equal.png', x=x + 656, y=y + 26, w=65, h=46, frames=3, z=91, delay=50, process=btn_anim_process, click=true};
      DS {'btn_semicolon', 'img', 'images/terminal/keyboard/btn_semicolon.png', x=x + 549, y=y + 52, w=65, h=55, frames=3, z=79, delay=50, process=btn_anim_process, click=true};
      DS {'btn_comma', 'img', 'images/terminal/keyboard/btn_comma.png', x=x + 483, y=y + 86, w=62, h=59, frames=3, z=68, delay=50, process=btn_anim_process, click=true};
      DS {'btn_slash', 'img', 'images/terminal/keyboard/btn_slash.png', x=x + 534, y=y + 86, w=69, h=59, frames=3, z=69, delay=50, process=btn_anim_process, click=true};
      DS {'btn_backslash', 'img', 'images/terminal/keyboard/btn_backslash.png', x=x + 589, y=y + 86, w=74, h=59, frames=3, z=70, delay=50, process=btn_anim_process, click=true};
    end;
  end;
  ondecor = function(s, name, press, x, y, btn) 
    if not press then
      if name == 'btn_shift' then
        if keys:state('shift') then
          keys:event(false, 'shift');
          D(name).frame_nr = 0;
          decor.dirty = true;
        else
          keys:event(true, 'shift');
          D(name).animated = 1;
          sounds[key_sounds[rnd(#key_sounds)]]:play();
        end;
        s:shift_buttons();
        return false;
      end;

      s:handledecor(true, mapping_decor_key[name]);
      return;
    end;

    return false;
  end;
  themeDispose = terminal_untheme;
  onkey = function(s, press, key)
    if not press and not release_handled_keys[key] then
      s._keyts = 0;
      return false;
    end;

    local ticks = instead.ticks();
    
    if s._key == key and ticks - s._keyts < 100 and s._keypress == press or key == 'return' and keys:state('alt') then
      return false;
    end;

    s._key = key;
    s._keyts = ticks;
    s._keypress = press;
    
    if key == 'left shift' then
      if press then
        if D('btn_shift').frame_nr ~= 2 then
          D('btn_shift').frame_nr = 2;
          sounds[key_sounds[rnd(#key_sounds)]]:play();
        end;
      else
        D('btn_shift').frame_nr = 0;
        D('btn_shift').animated = false;
      end;
      decor.dirty = true;
      s:shift_buttons();
      return;
    end;
    
    if not prefs.text_input or key == 'backspace' or key == 'left' or key == 'space' or key == 'return' then
      if keys:state('shift') and keymapping[key] then
        key = keymapping[key];
      end;
      
      s:oninput(key);
    end;
  end;
  handledecor = function(s, press, key)
    if key == 'left shift' then
      if press then
        if D('btn_shift').frame_nr ~= 2 then
          D('btn_shift').frame_nr = 2;
          sounds[key_sounds[rnd(#key_sounds)]]:play();
        end;
      else
        D('btn_shift').frame_nr = 0;
        D('btn_shift').animated = false;
      end;
      decor.dirty = true;
      s:shift_buttons();
      return;
    end;
    
    if keys:state('shift') and keymapping[key] then
      key = keymapping[key];
    end;
    
    s:oninput(key);
  end;
  ontext = function(s, char)
    if prefs.text_input then
      if (keymapping[char] or reverse_keymapping[char]) then
        s:oninput(char);
      elseif char ~= ' ' then
        sounds['error_beep']:play();
      end;
    end;
  end;
  oninput = function(s, key)
    sounds[key_sounds[rnd(#key_sounds)]]:play();
    
    local name = back_mapping_decor_key[key];
    if not name then
      local revkey = key;
      if reverse_keymapping[key] then
        revkey = reverse_keymapping[key];
      elseif keymapping[key] then
        revkey = keymapping[key];
      end;
      
      name = back_mapping_decor_key[revkey];
    end;
    
    if name then
      D(name).animated = 1;
      D(name).__delay = instead.ticks() - D(name).delay;
    end;
  
    if key == 'backspace' or key == 'left' then
      s.entry = s.entry:sub(1, #s.entry - 1);
    elseif key == 'right' or key == 'up' or key == 'down' then
    elseif key == 'space' then
      if #s.entry + #s.path + 3 > 63 then
        sounds['error_beep']:play();
        return;
      end;
      s.entry = s.entry .. ' ';
    elseif key == 'return' then
      local r = false;
      
      if s.command_preprocessor then
        r = s:command_preprocessor(s.entry);
      end;
      
      if not r then
        r = s:process_command(s.entry)
        if s:call_command(r[1], r[2]) then
          return;
        end;
        s.entry = '';
      end;
    else
      if #s.entry + #s.path + 3 > 63 then
        sounds['error_beep']:play();
        return;
      end;
    
      s.entry = s.entry .. key;
    end;
    if here() ^ s then
      local d = getscreen(false);
      if s.query_formatter then
        d:replace_line(s:query_formatter(s.path, s.entry));
      else
        d:replace_line(s.path .. '>' .. s.entry .. '_');
      end;
    end;
  end;
  onboot = function(s)
    s:process_directory_structure(s.directory_structure['C:']);
    
    s:makekeyboard();
    local d = getscreen(false);
    d:replace_line(s.path .. '>' .. s.entry .. '_');
  end;
}
