include "terminal/base"

room {
  nam = 'terminal_boot';
  disp = '';
  nosave = true;
  noautosave = true;
  nofollow = true;
  noinv = true;
  nofading = true;
  network = false;
  graphical = false;
  theme = 'terminal';
  onenter = function(s)
    if not s.network then
      pn(translate_now("terminal_locale", 'boot_from') .. ' ' .. translate_now("terminal_locale", 'no_device'));
      return false;
    else
      walk('terminal');
    end;
  end;
  themeInit = function(s)
    load_sounds('error_beep', 'pc_startup', 'keyboard1', 'keyboard2', 'keyboard3', 'keyboard4', 'keyboard5');
  end;
};

local iface_cmd=iface.cmd
function iface:cmd(inp,s)
  local a=std.split(inp)
  if a[1] == '@metaparser' and here().onmp then
    inp=inp:sub(#a[1]+2);
    inp=inp:sub(2,#inp-1);
    return here():onmp(inp);
  end;
  return iface_cmd(self, inp);
end;

terminal {
  {
    msg = '';
  };
  nam = 'terminal';
  timer = function(s)
    if s.status == 'wait_format' then
      if s.format_percent < 100 then
        s.format_percent = s.format_percent + 1;
      else
        s.status = 'format_done';
        
      end;
    elseif s.status == 'format_done' then
      --[[if s.format_percent < 120 then
        p(translate_now("terminal_locale", 'critical_failure'));
      end;]]
      
      s.format_percent = s.format_percent + 1;
      if s.format_percent > 150 then
        timer:stop();
        walk('terminal_boot_failure');
        return;
      end;
    end;
  end;
  decor = function(s)
    if s.status == 'command' then
      return s.msg .. '^' .. s.path .. '>';
    elseif s.status == 'crypt' or s.status == 'connect' then
      return s.msg .. '^' .. translate_now("terminal_locale", 'query_password');
    elseif s.status == 'format' then
      return s.msg .. '^' .. translate_now("terminal_locale", 'query_type_yes');
    elseif s.status == 'wait_format' then
      return s.msg .. '^' .. translate_now("terminal_locale", 'query_formatting') .. " " .. tostring(s.format_percent) .. '%';
    elseif s.status == 'format_done' then
      return s.msg .. '^' .. translate_now("terminal_locale", 'formatting_complete');
    end;
    return s.msg;
  end;
  push_line = function(s, line)
    s.msg = s.msg .. '^' .. line;
  end;
  makekeyboard = function(s)
  end;
  onmp = function(s, text)
    s.msg = '';
    local r = false;
      
    if s.command_preprocessor then
      r = s:command_preprocessor(text);
    end;
    
    if not here()^'terminal' then
        return iface_cmd(iface, "look");
      end;
    
    if not r then
      r = s:process_command(text)
      if s:call_command(r[1], r[2]) then
        if not here()^'terminal' then
          return iface_cmd(iface, "look");
        end;
        return;
      end;
    end;
    
    return s:decor();
  end;
  onenter = function(s)
    s.msg = '';
  end;
  way = {
    path {'#fail', translate("common_locale", 'continue'), 'terminal_boot_failure'}: disable();
  };
}

room {
  nam = 'terminal_boot_failure';
  theme = 'terminal';
  nofollow = true;
  noinv = true;
  disp = '';
  decor = function(s)
    pn(translate_now("terminal_locale", 'formatting_complete'));
    pn(translate_now("terminal_locale", 'critical_failure'));
  end;
  way = {
    path {translate("common_locale", 'continue'), 'ending_formatted'};
  };
}